from copy import deepcopy
from functools import partial

import numpy as np
from sklearn.preprocessing import normalize

from ml_recsys_tools.recommenders.recommender_base import BaseDFSparseRecommender
from ml_recsys_tools.utils.parallelism import map_batches_multiproc
from ml_recsys_tools.utils.similarity import top_N_sorted_on_sparse, top_N_sorted
from ml_recsys_tools.utils.instrumentation import log_time_and_shape


@log_time_and_shape
def interactions_mat_to_cooccurrence_mat(
        obs_mat, normalize_items=True, degree=1, base_min_cooccurrence=1,
        prune_ratio=0.5, decay=0.3, min_cooccurrence=3, trans_func='ones'):
    def prune_mat(m, ratio=0.0, cutoff=0):
        if (ratio == 0.0 and cutoff == 0) or (not len(m.data)):
            return m
        else:
            if ratio > 0.0:
                if len(m.data) > 50000:
                    data_sample = np.random.choice(
                        m.data, min(len(m.data), 10000), replace=False)
                else:
                    data_sample = m.data
                cutoff = max(cutoff, np.percentile(data_sample, int(100 * ratio)))

            m.data[m.data < cutoff] *= 0
            m.eliminate_zeros()
        return m

    def first_degree_cooccurrence(mat, min_cooc=1):

        if trans_func == 'ones':
            # binarize interaction
            mat.data = np.ones(mat.data.shape)

        elif trans_func == 'log':
            mat.data = np.log10(mat.data + 1)

        elif trans_func == 'none':
            pass

        else:
            raise ValueError('Unknown trans_func: %s' % trans_func)

        # 1st degree interaction matrix
        cooc_mat = mat.T * mat
        # remove self similarities
        cooc_mat.setdiag(0)
        # maybe less memory
        cooc_mat = cooc_mat.astype(np.float32)

        if min_cooc > 1:
            # threshold interactions
            cooc_mat = prune_mat(cooc_mat, 0.0, min_cooc)

        return cooc_mat

    cooc_mat_base = first_degree_cooccurrence(obs_mat, base_min_cooccurrence)

    if degree > 1:

        # keep weight constant
        total_weight = np.sum(cooc_mat_base.data)

        higher_deg_cooc = prune_mat(cooc_mat_base.copy(), prune_ratio, min_cooccurrence)

        for i in range(degree - 1):
            higher_deg_cooc += \
                decay ** (i + 1) * \
                higher_deg_cooc * \
                higher_deg_cooc

            # remove self similarities
            higher_deg_cooc.setdiag(0)

        higher_deg_cooc.data *= total_weight / (np.sum(higher_deg_cooc.data) + 1)  # avoid divide by 0

        cooc_mat_base += higher_deg_cooc

    # mormalization
    if normalize_items:
        cooc_mat_base = normalize(cooc_mat_base, norm='l1', axis=1)

    return cooc_mat_base


class BaseSimilarityRecommeder(BaseDFSparseRecommender):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.similarity_mat = None

    def _prep_for_fit(self, train_obs, **fit_params):
        self._set_fit_params(fit_params)
        self._set_data(train_obs)

    def _check_no_negatives(self):
        # prevents negative scores from being smaller than sparse zeros (e.g. for euclidean similarity)
        if len(self.similarity_mat.data) and np.min(self.similarity_mat.data) < 0.01:
            self.similarity_mat.data += np.abs(np.min(self.similarity_mat.data) - 0.01)

    def _get_recommendations_flat(
            self, user_ids, item_ids=None, exclude_training=True,
            n_rec=100, **kwargs):

        self._check_no_negatives()

        user_inds = self.sparse_mat_builder.uid_encoder.transform(user_ids)

        if item_ids is not None:
            item_inds = self.sparse_mat_builder.iid_encoder.transform(item_ids)
            item_inds.sort()
        else:
            item_inds = None

        def best_for_batch(u_inds):
            user_pred_mat = self.train_mat[u_inds, :] * self.similarity_mat
            if exclude_training:
                user_pred_mat -= self.train_mat[u_inds, :] * np.inf
            if item_inds is not None:
                user_pred_mat = user_pred_mat[:, item_inds]
            return top_N_sorted(user_pred_mat.toarray(), n_rec)

        ret = map_batches_multiproc(best_for_batch, user_inds, chunksize=200)

        best_inds = np.concatenate([r[0] for r in ret], axis=0)
        best_scores = np.concatenate([r[1] for r in ret], axis=0)

        # back to ids
        if item_inds is not None:
            best_inds = item_inds[best_inds.astype(int)]

        best_ids = self.sparse_mat_builder.iid_encoder.\
            inverse_transform(best_inds.astype(int))

        return self._format_results_df(
            source_vec=user_ids, target_ids_mat=best_ids, scores_mat=best_scores,
            results_format='recommendations_flat')

    def get_similar_items(self, item_ids=None, target_item_ids=None,
                          n_simil=10, results_format='lists', **kwargs):

        item_ids, target_item_ids = \
            self._check_item_ids_args(item_ids, target_item_ids)

        self._check_no_negatives()

        best_ids, best_scores = top_N_sorted_on_sparse(
            source_ids=item_ids,
            target_ids=target_item_ids,
            encoder=self.sparse_mat_builder.iid_encoder,
            sparse_mat=self.similarity_mat,
            n_top=n_simil,
        )

        simil_df = self._format_results_df(
            item_ids, target_ids_mat=best_ids,
            scores_mat=best_scores, results_format='similarities_' + results_format)
        return simil_df


class ItemCoocRecommender(BaseSimilarityRecommeder):

    default_fit_params = dict(
        normalize_items=True,
        degree=1,
        prune_ratio=0.0,
        decay=0.5,
        min_cooccurrence=1,
        base_min_cooccurrence=1,
        trans_func='ones',
    )

    def fit(self, train_obs, **fit_params):
        self._prep_for_fit(train_obs, **fit_params)
        self.similarity_mat = interactions_mat_to_cooccurrence_mat(
            self.train_mat, **self.fit_params)
        self.similarity_mat += self.similarity_mat.T


class UserCoocRecommender(ItemCoocRecommender):

    def fit(self, train_obs, **fit_params):
        self._prep_for_fit(train_obs, **fit_params)
        self.similarity_mat = interactions_mat_to_cooccurrence_mat(
            self.train_mat.T, **self.fit_params)

    def _get_recommendations_flat(
            self, user_ids, item_ids=None, exclude_training=True,
            n_rec=100, **kwargs):

        self._check_no_negatives()

        user_inds = self.sparse_mat_builder.uid_encoder.transform(user_ids)

        if item_ids is not None:
            item_inds = self.sparse_mat_builder.iid_encoder.transform(item_ids)
            item_inds.sort()
        else:
            item_inds = None

        def best_for_batch(u_inds):
            user_pred_mat = self.similarity_mat[u_inds, :] * self.train_mat
            if exclude_training:
                user_pred_mat -= self.train_mat[u_inds, :] * np.inf
            if item_inds is not None:
                user_pred_mat = user_pred_mat[:, item_inds]
            return top_N_sorted(user_pred_mat.toarray(), n_rec)

        ret = map_batches_multiproc(best_for_batch, user_inds, chunksize=1000)

        best_inds = np.concatenate([r[0] for r in ret], axis=0)
        best_scores = np.concatenate([r[1] for r in ret], axis=0)

        # back to ids
        if item_inds is not None:
            best_inds = item_inds[best_inds.astype(int)]

        best_ids = self.sparse_mat_builder.iid_encoder. \
            inverse_transform(best_inds.astype(int))

        return self._format_results_df(
            source_vec=user_ids, target_ids_mat=best_ids, scores_mat=best_scores,
            results_format='recommendations_flat')

    def get_similar_items(self, item_ids=None, target_item_ids=None, n_simil=10,
                          results_format='lists', **kwargs):
        raise NotImplementedError


class SimilarityDFRecommender(BaseSimilarityRecommeder):

    def get_similarity_builder(self):
        # this is hacky, but I want to make sure this recommender is even useful first
        # should be some other class' method
        simil_mat_builder = deepcopy(self.sparse_mat_builder)
        simil_mat_builder.uid_source_col = self._item_col_simil
        simil_mat_builder.iid_source_col = self._item_col
        simil_mat_builder.rating_source_col = self._prediction_col
        simil_mat_builder.n_rows = simil_mat_builder.n_cols
        simil_mat_builder.uid_encoder = simil_mat_builder.iid_encoder
        return simil_mat_builder

    def _prep_for_fit(self, train_obs, **fit_params):
        super()._prep_for_fit(train_obs, **fit_params)
        self.similarity_mat_builder = self.get_similarity_builder()

    def fit(self, train_obs, simil_df_flat, **fit_params):
        self._prep_for_fit(train_obs, **fit_params)
        self.similarity_mat = self.similarity_mat_builder. \
            build_sparse_interaction_matrix(simil_df_flat)
        self.similarity_mat += self.similarity_mat.T

    def continue_fit(self, simil_df_flat):
        partial_similarity_mat = self.similarity_mat_builder. \
            build_sparse_interaction_matrix(simil_df_flat)
        self.similarity_mat += partial_similarity_mat + partial_similarity_mat.T



import copy

import numpy as np
import pandas as pd
from sklearn.cluster import MiniBatchKMeans
import scipy.sparse as sp

from sklearn.preprocessing import LabelBinarizer, normalize, LabelEncoder
from sklearn_pandas import DataFrameMapper
from ml_recsys_tools.utils.sklearn_extenstions import NumericBinningBinarizer, GrayCodesNumericBinarizer

from ml_recsys_tools.data_handlers.interaction_handlers_base import ObservationsDF, RANDOM_STATE
from ml_recsys_tools.utils.instrumentation import LogCallsTimeAndOutput
from ml_recsys_tools.utils.geo import ItemsGeoMapper
from ml_recsys_tools.utils.logger import simple_logger as logger


class ExternalFeaturesDF(LogCallsTimeAndOutput):
    """
    handles external items features and feature engineering
    """

    _numeric_duplicate_suffix = '_num'

    def __init__(self, feat_df, id_col, num_cols=None, cat_cols=None, verbose=True):
        super().__init__(verbose)
        self.feat_df = feat_df.copy()
        self.id_col = id_col
        self.num_cols = num_cols
        self.cat_cols = cat_cols
        self._numeric_duplicate_cols = None
        self._feat_weight = None
        self.df_transformer = None
        if self.num_cols is None and self.cat_cols is None:
            self.infer_categorical_numerical_columns()

    def infer_categorical_numerical_columns(self,
                                            categorical_unique_ratio=0.05,
                                            categorical_n_unique=20):

        len_df = len(self.feat_df)

        if not len_df:
            raise ValueError('Features DF is empty')

        feat_cols = self.feat_df.columns.difference([self.id_col])

        self.num_cols, self.cat_cols = [], []

        for col in feat_cols:
            if str(self.feat_df[col].dtype) in ['O', 'o']:
                self.cat_cols.append(col)
            else:
                n_unique = self.feat_df[col].nunique()
                unique_ratio = n_unique / len_df
                if n_unique < categorical_n_unique or \
                        unique_ratio <= categorical_unique_ratio:
                    self.cat_cols.append(col)
                else:
                    self.num_cols.append(col)

    def apply_selection_filter(self, selection_filter=None):
        if selection_filter is not None and len(selection_filter) >= 1:
            # no selection applied for None, '', []
            self.cat_cols = [col for col in self.cat_cols if col in selection_filter]
            self.num_cols = [col for col in self.num_cols if col in selection_filter]

        self.feat_df = self.feat_df[[self.id_col] + self.cat_cols + self.num_cols]

        return self

    def _check_intersecting_num_and_cat_columns(self):
        self._numeric_duplicate_cols = list(set(self.cat_cols).intersection(set(self.num_cols)))
        if len(self._numeric_duplicate_cols):
            for col in self._numeric_duplicate_cols:
                alt_name = col + self._numeric_duplicate_suffix
                self.feat_df[alt_name] = self.feat_df[col].copy()
                self.num_cols.remove(col)
                self.num_cols.append(alt_name)

    def fit_transform_ids_df_to_mat(self,
                                    items_encoder,
                                    mode='binarize',
                                    normalize_output=False,
                                    add_identity_mat=False,
                                    numeric_n_bins=128,
                                    feat_weight=1.0):
        """
        creates a sparse feature matrix from item features

        :param items_encoder: the encoder that is used to filter and
            align the features dataframe to the sparse matrices
        :param mode: 'binarize' or 'encode'.
            'binarize' (default) - creates a binary matrix by label binarizing
            categorical and numeric feature.
            'encode' - only encodes the categorical features to integers and leaves numeric as is
        :param add_identity_mat: indicator whether to add a sparse identity matrix
            (as used when no features are used), as per LightFM's docs suggestion
        :param normalize_output:
            None (default) - no normalization
            'rows' - normalize rows with l1 norm
            anything else - normalize cols with l1 norm
        :param numeric_n_bins: number of bins for binning numeric features
        :param feat_weight:
            feature weight relative to identity matrix (can be used to emphasize one or the other)
            can also be a dictionary of weights to be applied to columns e.g. {'column_name': 10}

        :return: sparse feature mat n_items x n_features
        """

        self._check_intersecting_num_and_cat_columns()

        feat_df = self.feat_df[
            [self.id_col] + self.cat_cols + self.num_cols]
        # get only features for relevant items
        feat_df = feat_df[feat_df[self.id_col].isin(items_encoder.classes_)]
        # convert from id to index
        feat_df['item_ind'] = items_encoder.transform(feat_df[self.id_col])

        # reorder in correct index order
        n_items = len(items_encoder.classes_)
        full_feat_df = pd.merge(
            pd.DataFrame({'item_ind': np.arange(n_items)}),
            feat_df.drop([self.id_col], axis=1), on='item_ind', how='left'). \
            drop_duplicates('item_ind'). \
            set_index('item_ind', drop=True)

        # remove nans resulting form join
        # https://stackoverflow.com/questions/34913590/fillna-in-multiple-columns-in-place-in-python-pandas
        full_feat_df = full_feat_df.apply(lambda x: x.fillna(0) if x.dtype.kind in 'biufc' else x.fillna('.'))

        full_feat_df[self.cat_cols] = \
            full_feat_df[self.cat_cols].astype(str)

        if len(full_feat_df):
            self.df_transformer = self.init_df_transformer(
                mode=mode,
                categorical_feat_cols=self.cat_cols,
                numeric_feat_cols=self.num_cols,
                numeric_n_bins=numeric_n_bins)

            feat_mat = self.df_transformer.fit_transform(full_feat_df)

            if sp.issparse(feat_mat):
                feat_mat.eliminate_zeros()

            # weight the features before adding the identity mat
            self._feat_weight = feat_weight
            feat_mat = self._apply_weights_to_matrix(feat_mat)

            # normalize each row
            if normalize_output:
                axis = int(normalize_output == 'rows')
                feat_mat = normalize(feat_mat, norm='l1', axis=axis, copy=False)

            if add_identity_mat:
                # identity mat
                id_mat = sp.identity(n_items, dtype=np.float32, format='csr')

                assert sp.issparse(feat_mat), 'Trying to add identity mat to non-sparse matrix'

                full_feat_mat = self.concatenate_csc_matrices_by_columns(
                    feat_mat.tocsc(), id_mat.tocsc()).tocsr()
            else:
                full_feat_mat = feat_mat

            return full_feat_mat

        else:
            return None

    def transform_df_to_mat(self, feat_df):
        if self._numeric_duplicate_cols is not None:
            for col in self._numeric_duplicate_cols:
                feat_df[col + self._numeric_duplicate_suffix] = feat_df[col].copy()
        feat_df[self.cat_cols] = feat_df[self.cat_cols].astype(str)
        trans_mat = self.df_transformer.transform(feat_df)
        return self._apply_weights_to_matrix(trans_mat)

    def _apply_weights_to_matrix(self, feat_mat):
        if np.isscalar(self._feat_weight):
            feat_mat = feat_mat.astype(np.float32) * self._feat_weight
        elif isinstance(self._feat_weight, dict):
            for col, weight in self._feat_weight.items():
                cols_mask = np.core.defchararray.startswith(
                    self.df_transformer.transformed_names_, col)
                feat_mat[:, cols_mask] *= weight
        else:
            raise ValueError('Uknown feature weight format.')
        return feat_mat

    @staticmethod
    def concatenate_csc_matrices_by_columns(matrix1, matrix2):
        # https://stackoverflow.com/a/33259578/6485667
        new_data = np.concatenate((matrix1.data, matrix2.data))
        new_indices = np.concatenate((matrix1.indices, matrix2.indices))
        new_ind_ptr = matrix2.indptr + len(matrix1.data)
        new_ind_ptr = new_ind_ptr[1:]
        new_ind_ptr = np.concatenate((matrix1.indptr, new_ind_ptr))

        return sp.csc_matrix((new_data, new_indices, new_ind_ptr), dtype=np.float32)

    @staticmethod
    def init_df_transformer(mode, categorical_feat_cols, numeric_feat_cols,
                            numeric_n_bins=64):
        if mode=='binarize':
            feat_mapper = DataFrameMapper(
                [(cat_col,
                  LabelBinarizer(sparse_output=True))
                 for cat_col in categorical_feat_cols] +
                [(num_col,
                  NumericBinningBinarizer(n_bins=numeric_n_bins, sparse_output=True))
                  # GrayCodesNumericBinarizer(n_bins=numeric_n_bins, sparse_output=True))
                 for num_col in numeric_feat_cols],
                sparse=True
            )
        elif mode=='encode':
            feat_mapper = DataFrameMapper(
                [(cat_col,
                  LabelEncoder())
                 for cat_col in categorical_feat_cols],
                default = None  # pass other columns as is
            )
        else:
            raise NotImplementedError('Unknown transform mode')
        return feat_mapper


class ObsWithFeatures(ObservationsDF):

    def __init__(self, df_obs, df_items, items_id_col='item_id', **kwargs):
        super().__init__(df_obs=df_obs, **kwargs)
        self.item_id_col = items_id_col
        self.cluster_label_col = 'cluster_label'
        self.df_items = self._preprocess_items_df(df_items)
        self._filter_relevant_obs_and_items()

    def __add__(self, other):
        super().__add__(other)
        self.df_items = pd.concat([self.df_items, other.df_items])
        self.df_items.drop_duplicates(self.item_id_col, inplace=True)
        return self

    def __repr__(self):
        return super().__repr__() + ', %d Items' % len(self.df_items)

    def _preprocess_items_df(self, df_items):
        # make sure the ID col is of object type
        df_items[self.item_id_col] = df_items[self.item_id_col].astype(str)
        df_items.drop_duplicates(self.item_id_col, inplace=True)
        return df_items

    def _filter_relevant_obs_and_items(self):
        items_ids = self.df_items[self.item_id_col].unique()
        obs_ids = self.df_obs[self.iid_col].unique()
        self.df_obs = self.df_obs[self.df_obs[self.iid_col].isin(items_ids)].copy()
        self.df_items = self.df_items[self.df_items[self.item_id_col].isin(obs_ids)].copy()

    def filter_by_cluster_label(self, label):
        assert self.cluster_label_col in self.df_items.columns
        other = copy.deepcopy(self)
        other.df_items = self.df_items[self.df_items[self.cluster_label_col] == label].copy()
        other._filter_relevant_obs_and_items()
        return other

    def _apply_filter(self, mask_filter):
        other = copy.deepcopy(self)
        other.df_items = self.df_items[mask_filter].copy()
        other._filter_relevant_obs_and_items()
        return other

    def sample_observations(self,
                            n_users=None,
                            n_items=None,
                            method='random',
                            min_user_hist=0,
                            min_item_hist=0,
                            random_state=None):
        sample_df = super().sample_observations(n_users=n_users,
                                                n_items=n_items,
                                                method=method,
                                                min_user_hist=min_user_hist,
                                                min_item_hist=min_item_hist,
                                                random_state=random_state)
        other = copy.deepcopy(self)
        other.df_obs = sample_df.df_obs
        other._filter_relevant_obs_and_items()
        return other

    def filter_columns_by_df(self, other_df_obs):
        """
        removes users or items that are not in the other user dataframe
        :param other_df_obs: other dataframe, that has the same structure (column names)
        :return: new observation handler
        """
        other = super().filter_columns_by_df(other_df_obs)
        other._filter_relevant_obs_and_items()
        return other

    def remove_interactions_by_df(self, other_df_obs):
        other = super().remove_interactions_by_df(other_df_obs)
        other._filter_relevant_obs_and_items()
        return other

    def items_filtered_by_ids(self, item_ids):
        return self.df_items[self.df_items[self.item_id_col].isin(item_ids)]

    def get_items_df_for_user(self, user):
        item_ids = self.user_filtered_df(user)[self.iid_col].unique().tolist()
        df_items_view = self.items_filtered_by_ids(item_ids)
        return df_items_view

    def get_item_features_for_obs(self,
                                  categorical_unique_ratio=0.05,
                                  categorical_n_unique=20,
                                  selection_filter=None, **kwargs):
        """

        :param categorical_n_unique: consider categorical if less unique values than this
        :param categorical_unique_ratio: consider categorical if ratio of uniques to length less than this
        :param selection_filter: include only those column, if None or empty includes all
        :param kwargs:
        :return: dataframe of features, list of numeric columns, list of categorical columns
        """
        feat_df = self.df_items

        ext_feat = ExternalFeaturesDF(
            feat_df=feat_df, id_col=self.item_id_col)

        ext_feat.apply_selection_filter(selection_filter)

        return ext_feat


class ObsWithGeoFeatures(ObsWithFeatures):

    def __init__(self, df_obs, df_items, lat_col='lat', long_col='long', **kwargs):
        super().__init__(df_obs=df_obs, df_items=df_items, **kwargs)
        self.lat_col = lat_col
        self.long_col = long_col
        self.df_items = self._preprocess_geo_cols(self.df_items)

    @property
    def geo_cols(self):
        return [self.lat_col, self.long_col]

    def _preprocess_geo_cols(self, df_items):
        # remove nans
        filt_nan = df_items[self.lat_col].notnull() & \
                   ~df_items[self.lat_col].isin(['None']) & \
                   ~df_items[self.lat_col].isin(['nan'])
        df_no_nans = df_items[filt_nan].copy()

        # to float
        df_no_nans[self.lat_col] = df_no_nans[self.lat_col].astype(float)
        df_no_nans[self.long_col] = df_no_nans[self.long_col].astype(float)

        return df_no_nans

    def filter_by_location_range(self, min_lat, max_lat, min_long, max_long):
        """ e.g.
        min_lat = -33.851674299999999
        max_lat = -33.767248700000003
        min_long = 151.09386979999999
        max_long = 151.31997699999999
        """

        geo_filt = (self.df_items[self.lat_col] <= max_lat) & (self.df_items[self.lat_col] >= min_lat) & \
                   (self.df_items[self.long_col] <= max_long) & (self.df_items[self.long_col] >= min_long)

        return self._apply_filter(geo_filt)

    def filter_by_location_circle(self, center_lat, center_long, degree_dist):

        geo_filt = ((self.df_items[self.lat_col] - center_lat) ** 2 +
                    (self.df_items[self.long_col] - center_long) ** 2) <= degree_dist

        return self._apply_filter(geo_filt)

    def filter_by_location_square(self, center_lat, center_long, degree_side):
        offset = degree_side / 2
        return self.filter_by_location_range(
            center_lat - offset, center_lat + offset, center_long - offset, center_long + offset)

    def filter_by_location_rectangle(self, center_lat, center_long, lat_side, long_side):
        offset_lat = lat_side / 2
        offset_long = long_side / 2
        return self.filter_by_location_range(
            center_lat - offset_lat, center_lat + offset_lat,
            center_long - offset_long, center_long + offset_long)

    def geo_cluster_items(self, n_clusters=20):
        cls = MiniBatchKMeans(n_clusters=n_clusters,
                              random_state=RANDOM_STATE). \
            fit(self.df_items[self.geo_cols].as_matrix())

        self.df_items[self.cluster_label_col] = cls.labels_

    def calcluate_equidense_geo_grid(self, n_lat, n_long, overlap_margin, geo_box):

        df_items = self.filter_by_location_range(**geo_box).df_items

        geo_filters = []
        lat_bins = np.percentile(df_items.lat, np.linspace(0, 100, n_lat + 1))

        for i in range(n_lat):
            cur_lat_bins = lat_bins[i:(i + 2)]

            prop_slice = df_items[(df_items.lat <= cur_lat_bins[1]) & (df_items.lat >= cur_lat_bins[0])]

            long_bins = np.percentile(prop_slice.long, np.linspace(0, 100, n_long + 1))

            lat_center = np.mean(cur_lat_bins)
            side_lat = (cur_lat_bins[1] - cur_lat_bins[0]) + overlap_margin * 2

            for j in range(n_long):
                long_center = np.mean(long_bins[j:(j + 2)])
                side_long = (long_bins[j + 1] - long_bins[j]) + overlap_margin * 2
                geo_filters.append((lat_center, long_center, side_lat, side_long))

        return geo_filters

    @staticmethod
    def calcluate_simple_geo_grid(n_lat, n_long, overlap_margin, geo_box):
        # ranges
        lat_range = geo_box['max_lat'] - geo_box['min_lat']
        long_range = geo_box['max_long'] - geo_box['min_long']

        # distance from center to borders without overlap
        d_lat = lat_range / (n_lat * 2)
        d_long = long_range / (n_long * 2)

        # center locations
        lat_centers = np.linspace(
            geo_box['min_lat'] + d_lat, geo_box['max_lat'] - d_lat, n_lat)
        long_centers = np.linspace(
            geo_box['min_long'] + d_long, geo_box['max_long'] - d_long, n_long)

        # add overlap to distances
        side_lat = 2 * (d_lat + overlap_margin)
        side_long = 2 * (d_long + overlap_margin)

        # create geo filters
        geo_filters = []
        for lat_center in lat_centers:
            for long_center in long_centers:
                geo_filters.append((lat_center, long_center, side_lat, side_long))
        return geo_filters

    def get_mapper(self):
        return ObsGeoFeatMapper(obs_handler=self, mapper_class=ItemsGeoMapper)


class ObsGeoFeatMapper(ObsWithGeoFeatures):

    def __init__(self, obs_handler, mapper_class=ItemsGeoMapper, **kwargs):
        self.mapper_class = mapper_class
        super().__init__(df_obs=obs_handler.df_obs, df_items=obs_handler.df_items, **kwargs)

    def map_items_for_user(self, user):
        return self.mapper_class(self.get_items_df_for_user(user))

    def map_items_by_common_items(self, item_id, default_marker_size=2):

        users = self.items_filtered_df(item_id)[self.uid_col].unique().tolist()

        items_dfs = [self.get_items_df_for_user(user) for user in users]

        # unite all data and get counts
        all_data = pd.concat(items_dfs)
        counts = all_data[self.item_id_col].value_counts()
        all_data = all_data.set_index(self.item_id_col).drop_duplicates()
        all_data['counts'] = counts
        all_data.reset_index(inplace=True)

        mapper = self.mapper_class(all_data)  # for view init

        # add maps for each user's history
        colors = iter(self.mapper_class.get_n_spaced_colors(len(items_dfs)))
        for df in items_dfs:
            color = next(colors)
            mapper.add_markers(df, color=color, size=default_marker_size)
            mapper.add_heatmap(df, color=color, sensitivity=1, opacity=0.4)

        # add common items
        common_items = all_data[all_data['counts'] > 1]
        sizes = list(map(int, np.sqrt(common_items['counts'].as_matrix()) + default_marker_size))
        mapper.add_markers(common_items, color='white', size=sizes)

        return mapper

    def map_cluster_labels(self, df_items=None, sample_n=1000):

        if df_items is None:
            df_items = self.df_items

        unique_labels = df_items[self.cluster_label_col].unique()

        items_dfs = [df_items[df_items[self.cluster_label_col] == label].sample(sample_n)
                     for label in unique_labels]

        all_data = pd.concat(items_dfs)

        mapper = self.mapper_class(all_data)

        colors = iter(self.mapper_class.get_n_spaced_colors(len(items_dfs)))

        for df in items_dfs:
            color = next(colors)
            mapper.add_heatmap(df, color=color, spread=50, sensitivity=3, opacity=0.3)

        return mapper

    def compare_similarity_results(self, item_id, items_lists, scores_lists=None, names=None, print_data=True):
        df_item_source = self.items_filtered_by_ids([item_id])
        items_dfs = [self.items_filtered_by_ids(l) for l in items_lists]

        # add variant and scores field
        if names is None:
            names = [str(i) for i in range(len(items_lists))]
        if scores_lists is None:
            scores_lists = [[5] * len(df) for df in items_dfs]
        all_lists = [df.assign(variant=name, score=scores[:len(df)])
                     for df, name, scores in
                     zip(items_dfs + [df_item_source],
                         names + ['source'], scores_lists + [[0]])]
        all_data = pd.concat(all_lists)

        # add counts
        all_data = all_data.join(
            all_data[self.item_id_col].
                value_counts().to_frame('count'),
            on=self.item_id_col)
        all_data = all_data.set_index(self.item_id_col)

        if print_data:
            logger.info('\n' + str(all_data))

        mapper = self.mapper_class(all_data)  # for view init

        if names is None:
            names = [''] * len(items_dfs)

        colors = self.mapper_class.get_n_spaced_colors(len(items_dfs))

        # poor man's legend
        logger.info("Poor man's legend: " + str(list(zip(colors, names))))

        for i in range(len(items_lists)):

            if scores_lists[i] is None:
                size = 5
            else:
                # filter simil scores only for those items that we have data for
                listings_with_data = items_dfs[i][self.item_id_col].values
                simil_scores = [score for listing_id, score in
                                zip(items_lists[i], scores_lists[i])
                                if listing_id in listings_with_data]
                size = 3 + (10 * np.array(simil_scores) ** 3).astype(np.int32)
                size = [int(el) for el in size]

            mapper.add_markers(items_dfs[i], size=size, color=colors[i])
            mapper.add_heatmap(items_dfs[i], sensitivity=1, opacity=0.4, spread=50, color=colors[i])

        mapper.add_markers(df_item_source, color='black', size=7)

        return mapper

    def map_similar_items(self, item_id, similar_items, similarity_scores=None):
        return self.compare_similarity_results(item_id, [similar_items], [similarity_scores])

    def map_recommendation_variants(
            self, train_items, test_items, recs_variants, colors, default_marker_size=5):
        df_train = self.items_filtered_by_ids(train_items)
        df_test = self.items_filtered_by_ids(test_items)
        rec_dfs = [self.items_filtered_by_ids(rec) for rec in recs_variants]

        all_data = pd.concat(rec_dfs + [df_train, df_test])
        all_data = all_data.set_index(self.item_id_col).drop_duplicates()

        mapper = self.mapper_class(all_data)  # for view init

        # add maps for each user's history
        for df, color in zip(rec_dfs, colors):
            mapper.add_markers(df, color=color, size=default_marker_size)
            mapper.add_heatmap(df, color=color, sensitivity=1, opacity=0.4, spread=50)

        if len(df_train):
            mapper.add_markers(df_train, color='gray', size=6)
            # mapper.add_heatmap(df_train, color='black', sensitivity=1, opacity=0.4)

        if len(df_test):
            mapper.add_markers(df_test, color='black', size=6)
            # mapper.add_heatmap(df_test, color=(250, 50, 50), sensitivity=1, opacity=0.4)

        return mapper



import os
import pprint
from functools import partial
from multiprocessing import Queue, Process
from threading import Thread
from types import SimpleNamespace

import numpy as np
import pandas as pd
import copy

import time
from matplotlib import pyplot
from sklearn.feature_selection import mutual_info_regression

from sklearn.metrics import recall_score, precision_score, f1_score
from sklearn.ensemble.voting_classifier import VotingClassifier

from sklearn.preprocessing import LabelEncoder
from skopt import Optimizer
from skopt.callbacks import VerboseCallback
from skopt.learning import RandomForestRegressor
from skopt.plots import plot_convergence
from skopt.optimizer import forest_minimize, gp_minimize, gbrt_minimize, dummy_minimize
from skopt.utils import dimensions_aslist, point_asdict
import skopt.space

from ml_recsys_tools.utils.instrumentation import \
    log_time_and_shape, collect_named_init_params, LogCallsTimeAndOutput
from ml_recsys_tools.utils.logger import simple_logger

# monkey patch print function
skopt.callbacks.print = simple_logger.info


class Integer(skopt.space.Integer):
    # changes the sampled values from numpy ints to python ints
    def rvs(self, n_samples=1, random_state=None):
        return [int(s) for s in super().rvs(n_samples=n_samples, random_state=random_state)]


class Real(skopt.space.Real):
    # changes the sampled values from numpy floats to python floats
    def rvs(self, n_samples=1, random_state=None):
        return [float(s) for s in super().rvs(n_samples=n_samples, random_state=random_state)]


Categorical = skopt.space.Categorical


class SearchSpaceGuess:

    Categorical = Categorical
    Real = Real
    Integer = Integer

    def __init__(self, arg):
        """
        constructs an instance with attributes from the dictionary of names and default values
        :param arg: either a dict of {name: default value} or and instance or a class for which to construct a guess
        """
        if isinstance(arg, dict):
            self.set_from_dict(arg)
        else:
            self.set_from_class(arg)

    def set_from_dict(self, dictionary):
        for key, value in dictionary.items():
            if isinstance(value, bool):
                variable = self.Categorical([value])
            elif isinstance(value, int):
                variable = self.Integer(value, value + 1)
            elif isinstance(value, float):
                variable = self.Real(value, value + 1)
            else:
                variable = self.Categorical([value])
            setattr(self, key, variable)
        return self

    def __repr__(self):
        return pprint.pformat(self.__dict__)

    def set_from_class(self, obj):
        """
        Guess a class search space from init params and construct an object with
        params as attributes and skopt variable distributions as values
        :param obj: the object from which to set the init search space
        :return: a SearchSpace object with params attributes named as the parameters and skopt variables as values
        """

        if isinstance(obj, type):
            cls = obj
        else:
            cls = type(obj)

        all_init_params = collect_named_init_params(cls)

        search_space = dict([el
                             for d in all_init_params.values()
                             for el in list(d.items())])

        return self.set_from_dict(search_space)


class BayesSearchHoldOut(LogCallsTimeAndOutput):

    target_loss_col = 'target_loss'
    time_taken_col = 'time_taken'

    def __init__(self, search_space, pipeline, loss, random_ratio=0.5,
                 plot_graph=True, interrupt_message_file=None,  n_parallel=1,
                 verbose=True, smooth_transition=False, **kwargs):
        self.search_space = search_space
        self.loss = loss
        self.pipeline = pipeline
        self.train_data = None
        self.validation_data = None
        self.callbacks = None
        self.random_ratio = random_ratio
        self.plot_graph = plot_graph
        self.smooth_transition = smooth_transition
        self.interrupt_message_file = interrupt_message_file
        self.n_parallel = n_parallel
        super().__init__(verbose=verbose, **kwargs)
        self.all_metrics = pd.DataFrame()

    def values_to_dict(self, values):
        return point_asdict(self.search_space, values)

    class PrintIfBestCB:
        def __init__(self, search_inst):
            # self.prev_result = np.inf
            self.search_inst = search_inst

        def __call__(self, result):
            best_result = result.fun
            cur_result = result.func_vals[-1]
            if best_result >= cur_result:
                # self.prev_result = cur_result
                simple_logger.info('best params, iteration %d' % len(result.func_vals))
            simple_logger.info('params for loss=%f:' % cur_result)
            values = result.x_iters[-1]
            params = self.search_inst.values_to_dict(values)
            simple_logger.info(params)

    def init_pipeline(self, values):
        for i, value in enumerate(values):
            if isinstance(value, np.int_):
                values[i] = int(value)
            elif isinstance(value, np.float_):
                values[i] = float(value)
        params_dict = self.values_to_dict(values)
        pipeline = copy.copy(self.pipeline)
        pipeline.set_params(**params_dict)
        return pipeline

    def _add_loss_and_time_to_report(self, loss, time_taken, report_df=None):
        if report_df is None:
            report_df = pd.DataFrame()
        report_df[self.target_loss_col] = loss
        report_df[self.time_taken_col] = time_taken
        return report_df

    def _fit_predict_loss(self, pipeline):
        start = time.time()
        pipeline.fit(self.train_data['x'], self.train_data['y'])
        y_pred = pipeline.predict(self.validation_data['x'])
        loss = self.loss(self.validation_data['y'], y_pred)
        return loss, self._add_loss_and_time_to_report(loss, time.time() - start)

    def objective_func(self, values):
        try:
            self._check_interrupt()
            pipeline = self.init_pipeline(values)
            loss, report_df = self._fit_predict_loss(pipeline)
            return loss, report_df
        except Exception as e:
            simple_logger.exception(e)
            simple_logger.error(values)
            return 1.0, self._add_loss_and_time_to_report(0, 0)

    def _init_optimizer(self, n_calls):
        return Optimizer(
            dimensions=dimensions_aslist(search_space=self.search_space),
            base_estimator=RandomForestRegressor(n_estimators=10),
            n_initial_points=int(n_calls * self.random_ratio),
            acq_func="EI",
            acq_optimizer="sampling",
            acq_optimizer_kwargs=dict(n_points=1000, n_jobs=-1),
            acq_func_kwargs=dict(xi=0.01, kappa=1.96))

    def _init_callbacks(self, n_calls):
        self.callbacks = [
            self.PrintIfBestCB(self),
            VerboseCallback(n_total=n_calls, n_random=int(n_calls * self.random_ratio))]

    def _eval_callbacks(self, result):
        [c(result) for c in self.callbacks]

    def _smooth_transition(self, optimizer, progress):
        # this should transition from random to BO
        if self.smooth_transition:
            optimizer.n_points = int(2000 ** progress)
        return optimizer

    def optimize(self, train_data, validation_data, n_calls):
        self.train_data = train_data
        self.validation_data = validation_data
        optimizer = self._init_optimizer(n_calls)
        self._init_callbacks(n_calls)

        config_q = Queue(maxsize=1)  # this is to keep the configs fresh
        results_q = Queue()

        def _config_putter(q):
            for i in range(n_calls):
                q.put(optimizer.ask())
            for i in range(self.n_parallel):
                q.put('end')

        jobs = [Thread(target=_config_putter, name='_config_putter', args=(config_q,))]

        if self.n_parallel == 1:
            jobs.append(Thread(target=self._parallel_worker, args=(config_q, results_q)))  # for easier debugging
        else:
            jobs.extend([Process(target=self._parallel_worker, args=(config_q, results_q))
                   for _ in range(self.n_parallel)])
        [j.start() for j in jobs]

        # get results from queue
        for i in range(n_calls):
            next_x, next_y, report_df = results_q.get()
            self._record_all_metrics(report_df=report_df, values=next_x)
            optimizer = self._smooth_transition(optimizer, i / n_calls)
            result = optimizer.tell(next_x, next_y)
            self._eval_callbacks(result)

        [j.join() for j in jobs]
        return self._format_and_plot_result(result)

    def _parallel_worker(self, q_in, q_out):
        while True:
            next_x = q_in.get()
            if next_x == 'end':
                break
            next_y, report_df = self.objective_func(next_x)
            q_out.put((next_x, next_y, report_df))

    def _format_and_plot_result(self, result):
        best_values = result.x
        best_params = self.values_to_dict(best_values)
        best_model = self.init_pipeline(best_values)

        if self.plot_graph:
            pyplot.figure()
            plot_convergence(result)
            pyplot.plot(result.func_vals)

        # return res_bo, best_params, best_model
        return SimpleNamespace(**{
            'optimizer': self,
            'report': self.best_results_summary(result, percentile=0),
            'mutual_info_loss': self.params_mutual_info(),
            'mutual_info_time': self.params_mutual_info(self.time_taken_col),
            'result': result,
            'best_params': best_params,
            'best_model': best_model})

    def _check_interrupt(self):
        if self.interrupt_message_file is not None \
                and os.path.exists(self.interrupt_message_file):

            with open(self.interrupt_message_file) as f:
                message = f.readline()

            if 'stop' in message:
                raise InterruptedError('interrupted by "stop" message in %s'
                                       % self.interrupt_message_file)
            elif 'pause' in message:
                simple_logger.warn('Paused by "pause" message in %s'
                            % self.interrupt_message_file)
                while 'pause' in message:
                    time.sleep(1)
                    with open(self.interrupt_message_file) as f:
                        message = f.readline()
                self._check_interrupt()

            elif 'update' in message:
                simple_logger.warn('Updating HP space due to "update" message in %s'
                            % self.interrupt_message_file)
                raise NotImplementedError('not yet implemented')

    def _record_all_metrics(self, values, report_df=None):
        # records the time and the other metrics
        if report_df is None:
            report_df = pd.DataFrame()
        params_dict = self.values_to_dict(values)
        report_df = report_df.assign(**params_dict)
        self.all_metrics = self.all_metrics.append(report_df)

    def best_results_summary(self, res_bo, percentile=95):
        return self.all_metrics. \
            reset_index(). \
            drop('index', axis=1). \
            sort_values(self.target_loss_col)

    def params_mutual_info(self, target_col=None):
        if target_col is None:
            target_col = self.target_loss_col
        mutual_info = {}
        target = self.all_metrics[target_col].values.reshape(-1, 1)
        for feat in self.search_space.keys():
            vec = self.all_metrics[feat].values.reshape(-1, 1)
            try:
                mutual_info[feat] = \
                    mutual_info_regression(vec, target)[0]
            except ValueError:  # categorical feature (string)
                mutual_info[feat] = mutual_info_regression(
                    LabelEncoder().fit_transform(
                        vec.astype(str)).reshape(-1, 1), target)[0]
        return pd.DataFrame([mutual_info])


def early_stopping_runner(
        score_func, check_point_func,
        epochs_start=0, epochs_max=200, epochs_step=10,
        stop_patience=10, decline_threshold=0.05,
        plot_graph=True):
    res_list = []
    max_score = 0
    decline_counter = 0
    cur_epoch = 0
    epochs_list = []
    max_epoch = 0
    # find optimal number of epochs on validation data
    while cur_epoch <= epochs_max:

        cur_step = epochs_start + epochs_step if cur_epoch == 0 else epochs_step

        simple_logger.info('Training epochs %d - %d.' %
                           (cur_epoch, cur_epoch + cur_step))

        cur_epoch += cur_step
        epochs_list.append(cur_epoch)

        cur_score = score_func(cur_epoch, cur_step)
        res_list.append(cur_score)

        # early stopping logic
        if max_score * (1 - decline_threshold) > cur_score:
            decline_counter += cur_step
            if decline_counter >= stop_patience:
                break
        else:
            decline_counter = 0

        if cur_score > max_score:
            max_score = cur_score
            max_epoch = cur_epoch
            check_point_func()

    # print logging info
    scores_str = ','.join(['%d%%(%d)' % (int(100 * s / max_score), e)
                           for s, e in zip(res_list, epochs_list)])

    simple_logger.info('Early stopping: stopped fit after %d '
                       'epochs (max validation score: %f (@%d), all scores: %s)'
                       % (cur_epoch, max_score, max_epoch, scores_str))

    if plot_graph:
        pyplot.figure()
        pyplot.plot(epochs_list, res_list)

    return max_epoch


class VotingEnsemble(VotingClassifier):
    """
    Adds caching for probabilities to the VotingClassifer to enable quick predictions with updated weights
    (functools lru_cache is no good because it expects immutable inputs)
    """

    def __init__(self, estimators, voting='hard', weights=None, n_jobs=1,
                 flatten_transform=None):
        super().__init__(estimators, voting, weights, n_jobs, flatten_transform)
        self._proba_cache = None
        self._x_cache = None

    def _collect_probas(self, x):
        if (self._proba_cache is None) or (self._x_cache is None) or \
                (not np.asarray(list(self._x_cache) == list(x)).all()):
            self._proba_cache = VotingClassifier._collect_probas(self, x)
            self._x_cache = x
        return self._proba_cache


class EnsembleTrainer:
    def __init__(self, models_losses, models_params, cutoff_ratio, pipeline, data_dict):
        """

        example usage:

        # res_bo is result of optimisation run (sklearn format)
        res_y = res_bo.func_vals
        res_x = res_bo.x_iters
        # hp_space is hyper_params space,
        # params are dictionaries of params for initialization of the pipeline
        params = [point_asdict(hp_space, x) for x in res_x]

        ens_selection = EnsembleTrainer(
            models_losses=res_y,
            models_params=params,
            cutoff_ratio=0.2,
            pipeline=pipe,
            data_dict=data_dict
        )

        # than weights can be optimizid in two ways - forward-backward, or bayesian optimisation
        ens, res_ens = ens_selection.optimize_weights_by_skopt(n_iter=1000)
        ens, traj = ens_selection.optimize_weights_by_forward_selection(n_iter=1000)

        """
        self.data_dict = data_dict
        self.pipeline = pipeline
        self.cutoff_ratio = cutoff_ratio
        self.models_losses = models_losses
        self.models_params = models_params

        self._candidates = self._select_candidates()
        self.ensemble = self._fit_initial_ensemble()

    def _select_candidates(self):
        # sort
        idx = np.argsort(self.models_losses)
        all_results = np.array(self.models_losses)[idx]
        all_params = np.array(self.models_params)[idx]

        # choose ensemble candidates
        cutoff = min(np.percentile(all_results, int(self.cutoff_ratio * 100)),
                     (1 - (1 - self.cutoff_ratio) * (1 - np.min(all_results))))

        # choose the params that satisfy the cutoff
        ens_params = [all_params[i] for i in range(len(all_params))
                      if (all_results[i] <= cutoff)]

        # instantiate the models
        candidates = [copy.deepcopy(self.pipeline.set_params(**ens_params[i]))
                      for i in range(len(ens_params))]

        return candidates

    def _fit_initial_ensemble(self):
        self._select_candidates()

        # define and fit initial ensemeble
        ensemble = VotingEnsemble([(str(i), self._candidates[i])
                                   for i in range(len(self._candidates))], voting='soft')

        ensemble.fit(self.data_dict['x_train'], self.data_dict['y_train'])

        return ensemble

    def _loss_function(self, weights):
        self.ensemble.weights = weights
        return 1 - np.mean(
            recall_score(self.data_dict['y_valid'],
                         self.ensemble.predict(self.data_dict['x_valid']),
                         average=None))

    def _initial_guess(self):
        return list(np.array([1] + list(np.zeros(len(self._candidates) - 1))))  # top model

    def optimize_weights_by_skopt(self, n_iter=100):
        # fit ensemble weights by bayesian optimisation (mostly random search though

        res_ens = forest_minimize(
            self._loss_function,
            dimensions=[Real(0.0, 1.0)] * len(self._candidates),
            n_calls=n_iter,
            x0=self._initial_guess(),
            n_random_starts=int(n_iter * 0.5),
            base_estimator="RF",
            n_points=1000
        )

        plot_convergence(res_ens)

        self.ensemble.weights = res_ens.x

        return self.ensemble, res_ens

    def optimize_weights_by_forward_selection(self, n_iter=1000):

        n_models = len(self._candidates)

        cur_weights = list(np.zeros(n_models))
        cur_loss = 1.0
        trajectory = [(cur_loss, 0, 0, 'f')]
        loss_calcs = 0

        # forward
        while loss_calcs <= n_iter:
            new_losses = []
            for ens_i in range(n_models):
                new_weights = cur_weights.copy()
                new_weights[ens_i] += 1
                new_losses.append(self._loss_function(new_weights))
                loss_calcs += 1

            idx_add = int(np.argmin(new_losses))
            new_loss = min(new_losses)

            if new_loss <= cur_loss:
                cur_weights[idx_add] += 1
                cur_loss = new_loss.copy()
                trajectory.append((new_loss, idx_add, loss_calcs, 'f'))

        # backward
        still_pruning = True
        while still_pruning:
            still_pruning = False
            for ens_i in range(n_models):
                new_weights = cur_weights.copy()
                new_weights[ens_i] = 0
                if sum(new_weights) == 0.0:
                    break

                new_loss = self._loss_function(new_weights)
                loss_calcs += 1

                if new_loss <= cur_loss:
                    cur_weights = new_weights
                    cur_loss = new_loss.copy()
                    trajectory.append((cur_loss, ens_i, loss_calcs, 'b'))
                    still_pruning = True

        self.ensemble.weights = cur_weights
        return self.ensemble, trajectory

# def train_ensemble_by_skopt(models_losses, models_params, cutoff_ratio,
#                             n_iter, pipeline, data_dict):
#     '''
#     trains a voting ensemble by creating one out of top models
#     and than optimizing its weights on a validation set
#
#     '''
#     def ensemble_objective_func(x, *args):
#         ens.weights = x
#         return 1 - np.mean(
#             recall_score(data_dict['y_valid'],
#                          ens.predict(data_dict['x_valid']),
#                          average=None))
#
#     # sort
#     idx = np.argsort(models_losses)
#     all_results = np.array(models_losses)[idx]
#     all_params = np.array(models_params)[idx]
#
#     # choose ensemble candidates
#     cutoff = min(np.percentile(all_results, int(cutoff_ratio * 100)),
#                  cutoff_ratio * (1 - np.min(all_results)))
#
#     # choose the params that satisfy the cutoff
#     ens_params = [all_params[i] for i in range(len(all_params))
#                   if (all_results[i] <= cutoff)]
#
#     # instantiate the models
#     ens_ests = [copy.deepcopy(pipeline.set_params(**ens_params[i]))
#                 for i in range(len(ens_params))]
#
#     # define and fit initial ensemeble
#     ens = VotingEnsemble([(str(i), ens_ests[i])
#                           for i in range(len(ens_ests))], voting='soft')
#
#     ens.fit(data_dict['x_train'], data_dict['y_train'])
#
#     # fit ensemble weights by bayesian optimisation (mostly random search though
#     initial_guess = np.array([1] + list(np.zeros(len(ens_ests) - 1)))  # top model
#     res_ens = forest_minimize(ensemble_objective_func,
#                               dimensions=[Real(0.0, 1.0)] * len(ens_ests),
#                               n_calls=n_iter,
#                               x0=list(initial_guess),
#                               n_random_starts=int(n_iter * 0.5),
#                               )
#
#     plot_convergence(res_ens)
#
#     ens.weights = res_ens.x
#
#     return ens, res_ens



"""
This is a basic example on datasets-1M demonstrating:
    - basic data ingestion without any item/user features
    - LightFM recommender
    - basic evaluation
    - recommendations output
    - similar items output
"""

# dataset: download and prepare dataframes
from ml_recsys_tools.datasets.prep_movielense_data import get_and_prep_data

rating_csv_path, users_csv_path, movies_csv_path = get_and_prep_data()

# read the interactions dataframe and create a data handler object and  split to train and test
import pandas as pd

ratings_df = pd.read_csv(rating_csv_path)
from ml_recsys_tools.data_handlers.interaction_handlers_base import ObservationsDF

obs = ObservationsDF(ratings_df, uid_col='userid', iid_col='itemid')
train_obs, test_obs = obs.split_train_test(ratio=0.2)

# train and test LightFM recommender
from ml_recsys_tools.recommenders.lightfm_recommender import LightFMRecommender

lfm_rec = LightFMRecommender()
lfm_rec.fit(train_obs, epochs=10)

# print summary evaluation report:
print(lfm_rec.eval_on_test_by_ranking(test_obs.df_obs, prefix='lfm ', n_rec=100))

# get all recommendations and print a sample (training interactions are filtered out by default)
recs = lfm_rec.get_recommendations(lfm_rec.all_users, n_rec=5)
print(recs.sample(5))

# get all similarities and print a sample
simils = lfm_rec.get_similar_items(lfm_rec.all_items, n_simil=5)
print(simils.sample(10))



import numpy as np
import pandas as pd
from itertools import islice
import multiprocessing
from multiprocessing.pool import ThreadPool, Pool

N_CPUS = multiprocessing.cpu_count()


def batch_generator(iterable, n=1):
    if hasattr(iterable, '__len__'):
        # https://stackoverflow.com/questions/8290397/how-to-split-an-iterable-in-constant-size-chunks
        l = len(iterable)
        for ndx in range(0, l, n):
            yield iterable[ndx:min(ndx + n, l)]

    elif hasattr(iterable, '__next__'):
        # https://stackoverflow.com/questions/1915170/split-a-generator-iterable-every-n-items-in-python-splitevery
        i = iter(iterable)
        piece = list(islice(i, n))
        while piece:
            yield piece
            piece = list(islice(i, n))
    else:
        raise ValueError('Iterable is not iterable?')


def map_batches_multiproc(func, iterable, chunksize, multiproc_mode='threads',
                          n_threads=None, threads_per_cpu=1.0):
    if n_threads is None:
        n_threads = int(threads_per_cpu * N_CPUS)

    if hasattr(iterable, '__len__') and len(iterable) <= chunksize:
        return [func(iterable)]

    with pool_type(multiproc_mode)(n_threads) as pool:
        batches = batch_generator(iterable, n=chunksize)
        return list(pool.imap(func, batches))


def pool_type(parallelism_type):
    if 'process' in parallelism_type.lower():
        return Pool
    elif 'thread' in parallelism_type.lower():
        return ThreadPool
    else:
        raise ValueError('Unsupported value for "parallelism_type"')


def parallelize_dataframe(df, func, n_partitions=N_CPUS, parallelism_type='process'):
    # with Pool(n_partitions) as pool:
    #     return pd.concat(pool.map(func, np.array_split(df, n_partitions)))
    df_split = np.array_split(df, n_partitions)
    with pool_type(parallelism_type)(n_partitions) as pool:
        res = pool.map(func, df_split)
    df = pd.concat(res)
    return df


def parallelize_array(arr, func, n_partitions=N_CPUS, parallelism_type='process'):
    # with Pool(n_partitions) as pool:
    #     return np.concatenate(pool.map(func, np.array_split(arr, n_partitions)))
    arr_split = np.array_split(arr, n_partitions)
    with pool_type(parallelism_type)(n_partitions) as pool:
        res = pool.map(func, arr_split)
    arr = np.concatenate(res)
    return arr



import pandas as pd
from pandas.compat import StringIO
from sklearn.metrics.classification import classification_report
import warnings
from functools import partial
from ml_recsys_tools.utils.parallelism import parallelize_dataframe


def console_settings():
    pd.set_option('display.max_colwidth', 300)
    pd.set_option('display.max_rows', None)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', 1000)


def hist_by_groups(groups):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(3, 7)
    ax = ax.ravel()
    for i, g in enumerate(groups):
        g[1].hist(label=g[0], alpha=1.0, ax=ax[i])
        ax[i].legend()
    return ax


def classification_report_df(y_true, y_pred):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        return pd.read_csv(StringIO(classification_report(y_true, y_pred)), sep=r"[ \t]{2,}")


# slow
def explode_df(df, cols):
    static_cols = [col for col in df.columns if col not in cols]
    rows = []
    for _, row in df.iterrows():
        stat_list = row[static_cols].tolist()
        for i in range(len(row[cols[0]])):
            new_row = []
            for col in cols:
                new_row.append(row[col][i])
            new_row = new_row + stat_list
            rows.append(new_row)
    return pd.DataFrame(rows, columns=cols + static_cols)


def explode_df_parallel(df, cols):
    return parallelize_dataframe(df, partial(explode_df, cols=cols))

# def explode_df_np(df, cols):
#     static_cols = [col for col in df.columns if col not in cols]
#     new_mats = []
#     for _, row in df.iterrows():
#         stat_arr = np.matrix(row[static_cols].tolist())
#         n_rep = len(row[cols[0]])
#         stat_mat = np.repeat(stat_arr, n_rep, axis=0)
#         rows_mat = np.concatenate((stat_mat, *tuple(np.matrix(row[col]).T for col in cols)), axis=1)
#         new_mats.append(rows_mat)
#
#     full_mat = np.concatenate(tuple(new_mats), axis=0)
#     return pd.DataFrame(full_mat, columns=static_cols+cols)


# def explode_df(df, cols, fill_value=''):
#     '''
#     from here:
#         https://stackoverflow.com/questions/45846765/efficient-way-to-unnest-explode-multiple-list-columns-in-a-pandas-dataframe
#     :param df:
#     :param cols:
#     :param fill_value:
#     :return:
#     '''
#     # make sure `lst_cols` is a list
#     if cols and not isinstance(cols, list):
#         cols = [cols]
#     # all columns except `lst_cols`
#     idx_cols = df.columns.difference(cols)
#
#     # calculate lengths of lists
#     lens = df[cols[0]].str.len()
#
#     if (lens > 0).all():
#         # ALL lists in cells aren't empty
#         return pd.DataFrame({
#             col:np.repeat(df[col].values, df[cols[0]].str.len())
#             for col in idx_cols
#         }).assign(**{col:np.concatenate(df[col].values) for col in cols}) \
#           .loc[:, df.columns]
#     else:
#         # at least one list in cells is empty
#         return pd.DataFrame({
#             col:np.repeat(df[col].values, df[cols[0]].str.len())
#             for col in idx_cols
#         }).assign(**{col:np.concatenate(df[col].values) for col in cols}) \
#           .append(df.loc[lens==0, idx_cols]).fillna(fill_value) \
#           .loc[:, df.columns]



import logging
import functools
import time
import inspect
from types import FunctionType
from abc import ABC, ABCMeta
from threading import Thread
from psutil import virtual_memory, cpu_percent

from ml_recsys_tools.utils.logger import simple_logger


class LoggingVerbosity:
    def __init__(self, verbose=True, min_time=1):
        """
        :param verbose: true or false - whether to print to logging.INFO
        :param min_time: minimum time for which to print, if function call is shorter - nothing is printed
        """
        self.verbose = verbose
        self.min_time = min_time

    @property
    def level(self):
        return logging.INFO if self.verbose else logging.DEBUG


LOGGING_VERBOSITY = LoggingVerbosity()


def variable_info(result):
    if hasattr(result, 'shape'):
        shape_str = 'shape: %s' % str(result.shape)
    elif isinstance(result, tuple) and len(result) <= 3:
        shape_str = 'tuple: (' + ','.join([variable_info(el) for el in result]) + ')'
    elif hasattr(result, '__len__'):
        shape_str = 'len: %s' % str(len(result))
    else:
        shape_str = str(result)[:50] + '...'

    ret_str = str(type(result)) + ', ' + shape_str
    return ret_str


def log_time_and_shape(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        sys_monitor = ResourceMonitor().start()

        start = time.time()

        result = fn(*args, **kwargs)

        elapsed = time.time() - start
        duration_str = '%.2f' % elapsed

        sys_monitor.stop()
        mem_str = 'mem: %s%%(peak:%s%%) cpu:%d%%' % \
                  (sys_monitor.current_memory, sys_monitor.peak_memory, int(sys_monitor.avg_cpu_load))

        ret_str = variable_info(result)

        stack_depth = get_stack_depth()

        fn_str = class_name(fn) + fn.__name__

        msg = ' ' * stack_depth + \
              '%s, elapsed: %s, returned: %s, sys %s' % \
              (fn_str, duration_str, ret_str, mem_str)

        if elapsed >= LOGGING_VERBOSITY.min_time:
            simple_logger.log(LOGGING_VERBOSITY.level, msg)

        return result

    return inner


def timer_deco(fn):
    log_format = 'function %s: %s s'

    @functools.wraps(fn)
    def inner(*args, **kwargs):
        start = time.time()
        result = fn(*args, **kwargs)
        duration = time.time() - start
        simple_logger.log(LOGGING_VERBOSITY.level, log_format,
                          fn.__name__, duration)
        return result

    return inner


class ResourceMonitor:
    def __init__(self, interval=0.2):
        self.interval = interval
        self._init_counters()
        self._thread = None
        self._run_condition = False

    def _init_counters(self):
        self.current_memory = 0
        self.peak_memory = 0
        self.avg_cpu_load = 0
        self._n_measurements = 0

    def __del__(self):
        if self._run_condition:
            self._run_condition = False
            self._thread.join(self.interval + 0.1)

    @staticmethod
    def _current():
        try:
            return virtual_memory().percent, cpu_percent()
        except KeyError:
            # for some reason there's a KeyError: ('psutil',) in psutil
            return 0, 0

    def _measure(self):
        cur_mem, cur_cpu = self._current()
        self.current_memory = cur_mem
        self.peak_memory = max(self.peak_memory, cur_mem)
        self.avg_cpu_load = (self.avg_cpu_load * self._n_measurements + cur_cpu) / \
                            (self._n_measurements + 1)
        self._n_measurements += 1

    def _thread_loop(self):
        while self._run_condition:
            self._measure()
            time.sleep(self.interval)

    def start(self):
        self._init_counters()
        if self._thread is not None:
            self._thread.join(0)
        self._run_condition = True
        self._thread = Thread(target=self._thread_loop, name='ResourceMonitor')
        self._thread.start()
        return self

    def stop(self):
        self._run_condition = False
        self._measure()


def get_stack_depth():
    try:
        return len(inspect.stack())
    except IndexError as e:
        # there is a bug in inspect module: https://github.com/ipython/ipython/issues/1456/
        return 0


def class_name(fn):
    cls = get_class_that_defined_method(fn)
    cls_str = cls.__name__ + '.' if cls else ''
    return cls_str


def get_class_that_defined_method(meth):
    # from https://stackoverflow.com/questions/3589311/
    # get-defining-class-of-unbound-method-object-in-python-3/25959545#25959545
    # modified to return first parent in reverse MRO
    if inspect.ismethod(meth):
        for cls in inspect.getmro(meth.__self__.__class__)[::-1]:
            if cls.__dict__.get(meth.__name__) is meth:
                return cls
        meth = meth.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0], None)
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', None)  # handle special descriptor objects


def collect_named_init_params(cls, skip_empty=True, ignore_classes=(object, ABC)):
    """
    a method to get all named params from all classed in this class' MRO
    can be used to infer the possible search space for hyperparam optimization

    :param skip_empty: whether to skip classes with no named parameters
    :return: a nested dict of {class-name: dict of {named-init-params: default values}}
    """
    params = {}
    for c in inspect.getmro(cls):
        if c not in ignore_classes:
            named_params = [
                p for p in inspect.signature(c.__init__).parameters.values()
                if p.name != 'self'
                   and p.kind != p.VAR_KEYWORD
                   and p.kind != p.VAR_POSITIONAL]

            if skip_empty and not named_params:
                continue

            params[c.__name__] = {
                p.name: p.default
                if p.default is not p.empty else ''
                for p in named_params}
    return params


# https://stackoverflow.com/questions/10067262/automatically-decorating-every-instance-method-in-a-class
# decorate all instance methods (unless excluded) with the same decorator
def decorate_all_metaclass(decorator):
    # check if an object should be decorated
    def do_decorate(attr, value):
        return ('__' not in attr and
                isinstance(value, (FunctionType, classmethod)) and
                getattr(value, 'decorate', True))

    class DecorateAll(ABCMeta):
        def __new__(cls, name, bases, dct):
            if dct.get('decorate', True):
                for attr, value in dct.items():
                    if do_decorate(attr, value):
                        if isinstance(value, classmethod):
                            dct[attr] = classmethod(decorator(value.__func__))
                        else:
                            dct[attr] = decorator(value)
            return super(DecorateAll, cls).__new__(cls, name, bases, dct)

        def __setattr__(self, attr, value):
            if do_decorate(attr, value):
                value = decorator(value)
            super(DecorateAll, self).__setattr__(attr, value)

    return DecorateAll


class LogCallsTimeAndOutput(metaclass=decorate_all_metaclass(log_time_and_shape)):

    def __init__(self, verbose, **kwargs):
        self.verbose = verbose

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, verbose):
        self._verbose = verbose
        if not self.verbose:
            self.decorate = False

    @property
    def logging_decorator(self):
        """
        this is for decorating inner scope functions
        :return: the logging decorator if verbose is True, empty decorator otherwise
        """
        if self.verbose:
            return log_time_and_shape
        else:
            return lambda f: f

    @staticmethod
    def do_not_decorate(f):
        """
        this is for excluding functions from decorating
        :return: the logging decorator if verbose is True, empty decorator otherwise
        """
        setattr(f, 'decorate', False)
        return f



"""
This is an example on datasets-1M demonstrating:
    - More advanced fitting features: fit, evaluation, early stopping, hyper-param search
"""

from ml_recsys_tools.datasets.prep_movielense_data import get_and_prep_data
import pandas as pd
from ml_recsys_tools.data_handlers.interaction_handlers_base import ObservationsDF
from ml_recsys_tools.recommenders.lightfm_recommender import LightFMRecommender

rating_csv_path, users_csv_path, movies_csv_path = get_and_prep_data()
ratings_df = pd.read_csv(rating_csv_path)

obs = ObservationsDF(ratings_df, uid_col='userid', iid_col='itemid')
train_obs, test_obs = obs.split_train_test(ratio=0.2)
lfm_rec = LightFMRecommender()


# train LightFM with early stopping and print evaluation results
lfm_rec.fit_with_early_stop(train_obs, epochs_max=30, epochs_step=1, stop_patience=1,
                            valid_ratio=0.2, metric='n-MRR@10', refit_on_all=True)
print(lfm_rec.eval_on_test_by_ranking(test_obs.df_obs, prefix='lfm early stop '))


# perform a hyperparameter search on LightFM recommender
space = lfm_rec.guess_search_space()
hp_results = lfm_rec.hyper_param_search(
    train_obs,
    metric='n-MRR@10',
    hp_space=dict(
        no_components=space.Integer(20, 100),
        epochs=space.Integer(5, 50),
        item_alpha=space.Real(1e-8, 1e-5, prior='log-uniform')
    ),
    n_iters=5,
)
print(hp_results.report)
# refit best model on full training data
# and print evaluation results
hp_results.best_model.fit(train_obs)
print(hp_results.best_model.eval_on_test_by_ranking(test_obs, prefix='lfm hp search '))



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='ml_recsys_tools',
    version='0.5.2.4',
    description='Tools for recommendation systems development',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/DomainGroupOSS/ml-recsys-tools',
    author='Domain group (Arthur Deygin)',
    author_email='arthurdgn@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords='recommendations machine learning',
    packages=find_packages(exclude=['tests', 'examples']),
    install_requires=['requests', 'scipy', 'numpy', 'pandas',
                      'scikit_learn', 'lightfm', 'implicit', 'psutil',
                      'scikit_optimize', 'gmaps', 'boto3', 'redis',
                      'sklearn_pandas', 'matplotlib'],
)

## rm -rf dist && python3 setup.py sdist bdist_wheel && twine upload dist/*



"""
Example explaining the peculiriaties of evaluation
"""

from ml_recsys_tools.datasets.prep_movielense_data import get_and_prep_data
import pandas as pd
from ml_recsys_tools.data_handlers.interaction_handlers_base import ObservationsDF
from ml_recsys_tools.recommenders.lightfm_recommender import LightFMRecommender

rating_csv_path, users_csv_path, movies_csv_path = get_and_prep_data()
ratings_df = pd.read_csv(rating_csv_path)

obs = ObservationsDF(ratings_df, uid_col='userid', iid_col='itemid')
train_obs, test_obs = obs.split_train_test(ratio=0.2)

# train and test LightFM recommender
lfm_rec = LightFMRecommender()
lfm_rec.fit(train_obs, epochs=10)

# print evaluation results:
# for LightFM there is an exact method that on large and sparse
# data might be too slow (for this data it's much faster though)
print(lfm_rec.eval_on_test_by_ranking_exact(test_obs.df_obs, prefix='lfm regular exact '))

# this ranking evaluation is done by sampling top n_rec recommendations
# rather than all ranks for all items (very slow and memory-wise expensive for large data).
# choosing higher values for n_rec makes
# the evaluation more accurate (less pessimmistic)
# this way the evaluation is mostly accurate for the top results,
# and is quite pessimmistic (especially for AUC, which scores for all ranks) and any non @k metric
print(lfm_rec.eval_on_test_by_ranking(test_obs.df_obs, prefix='lfm regular ', n_rec=100))
